/*
 * Clock and power management. This exposes the combined functionality
 * provided by the X1600 and related SoCs. The power management
 * functionality could be exposed using a separate driver.
 *
 * Copyright (C) 2017, 2018, 2020, 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-x1600.h"
#include <math.h>
#include <stdio.h>



// Register locations.

enum Regs : unsigned
{
  Clock_control       = 0x000,  // CPCCR
  Low_power_control   = 0x004,  // LCR
  Clock_gate0         = 0x020,  // CLKGR0
  Clock_gate1         = 0x028,  // CLKGR1
  Sleep_control       = 0x024,  // OPCR (oscillator and power control)
  Clock_status        = 0x0d4,  // CPCSR
  Ddr_divider         = 0x02c,  // DDRCDR
  Mac_divider         = 0x054,  // MACCDR
  I2s_divider0        = 0x060,  // I2SCDR
  I2s_divider1        = 0x070,  // I2S1CDR
  Lcd_divider         = 0x064,  // LPCDR
  Msc_divider0        = 0x068,  // MSC0CDR
  Msc_divider1        = 0x0a4,  // MSC1CDR
  Sfc_divider         = 0x074,  // SFCCDR
  Ssi_divider         = 0x05c,  // SSICDR
  Cim_divider         = 0x078,  // CIMCDR
  Pwm_divider         = 0x06c,  // PWMCDR
  Can_divider0        = 0x0a0,  // CAN0CDR
  Can_divider1        = 0x0a8,  // CAN1CDR
  Cdbus_divider       = 0x0ac,  // CDBUSCDR
  Macphy0_divider     = 0x0e4,  // MPHY0C
  Cpm_interrupt       = 0x0b0,  // CPM_INTR
  Cpm_interrupt_en    = 0x0b4,  // CPM_INTRE
  Cpm_swi             = 0x0bc,  // CPM_SFTINT
  Ddr_gate            = 0x0d0,  // DRCG
  Cpm_scratch_prot    = 0x038,  // CPSPPR
  Cpm_scratch         = 0x034,  // CPSPR
  Usb_param_control0  = 0x03c,  // USBPCR
  Usb_reset_detect    = 0x040,  // USBRDT
  Usb_vbus_jitter     = 0x044,  // USBVBFIL
  Usb_param_control1  = 0x048,  // USBPCR1
  Pll_control         = 0x00c,  // CPPCR
  Pll_control_A       = 0x010,  // CPAPCR
  Pll_control_M       = 0x014,  // CPMPCR
  Pll_control_E       = 0x018,  // CPEPCR
  Pll_fraction_A      = 0x084,  // CPAPACR
  Pll_fraction_M      = 0x088,  // CPMPACR
  Pll_fraction_E      = 0x08c,  // CPEPACR

  // Special value

  Reg_undefined       = 0xfff,
};

enum Clock_source_bits : unsigned
{
  // Clock_control

  Clock_source_main     = 30, // SEL_SRC (output to SCLK_A)
  Clock_source_cpu      = 28, // SEL_CPLL (output to CCLK)
  Clock_source_hclock0  = 26, // SEL_H0PLL (output to AHB0)
  Clock_source_hclock2  = 24, // SEL_H2PLL (output to AHB2)

  // Divider registers

  Clock_source_can0     = 30, // CA0CS
  Clock_source_can1     = 30, // CA1CS
  Clock_source_cdbus    = 30, // CDCS
  Clock_source_cim      = 30, // CIMPCS
  Clock_source_ddr      = 30, // DCS
  Clock_source_i2s      = 31, // I2PCS
  Clock_source_lcd      = 30, // LPCS
  Clock_source_mac      = 30, // MACPCS
  Clock_source_msc0     = 30, // MPCS
  Clock_source_msc1     = 30, // MPCS
  Clock_source_pwm      = 30, // PWMPCS
  Clock_source_sfc      = 30, // SFCS
  Clock_source_ssi      = 30, // SPCS

  // Special value

  Clock_source_undefined = 32,
};

enum Clock_source_values : unsigned
{
  Source_mME_main       = 0,
  Source_mME_pll_M      = 1,
  Source_mME_pll_E      = 2,

  // Special value

  Source_mask           = 0x3,
};

enum Clock_gate_bits : unsigned
{
  // Clock_control

  Clock_gate_main       = 23, // GATE_SCLKA

  // Clock_gate0

  Clock_gate_ddr        = 31, // DDR
  Clock_gate_ahb0       = 29, // AHB0
  Clock_gate_apb0       = 28, // APB0
  Clock_gate_rtc        = 27, // RTC
  Clock_gate_aes        = 24, // AES
  Clock_gate_lcd_pixel  = 23, // LCD
  Clock_gate_cim        = 22, // CIM
  Clock_gate_dma        = 21, // PDMA
  Clock_gate_ost        = 20, // OST
  Clock_gate_ssi0       = 19, // SSI0
  Clock_gate_timer      = 18, // TCU
  Clock_gate_dtrng      = 17, // DTRNG
  Clock_gate_uart2      = 16, // UART2
  Clock_gate_uart1      = 15, // UART1
  Clock_gate_uart0      = 14, // UART0
  Clock_gate_sadc       = 13, // SADC
  Clock_gate_audio      = 11, // AUDIO
  Clock_gate_ssi_slv    = 10, // SSI_SLV
  Clock_gate_i2c1       = 8,  // I2C1
  Clock_gate_i2c0       = 7,  // I2C0
  Clock_gate_msc1       = 5,  // MSC1
  Clock_gate_msc0       = 4,  // MSC0
  Clock_gate_otg        = 3,  // OTG
  Clock_gate_sfc        = 2,  // SFC
  Clock_gate_efuse      = 1,  // EFUSE
  Clock_gate_nemc       = 0,  // NEMC

  // Clock_gate1

  Clock_gate_arb        = 30, // ARB
  Clock_gate_mipi_csi   = 28, // MIPI_CSI
  Clock_gate_intc       = 26, // INTC
  Clock_gate_gmac0      = 23, // GMAC0
  Clock_gate_uart3      = 16, // UART3
  Clock_gate_i2s0_tx    = 9,  // I2S0_dev_tclk
  Clock_gate_i2s0_rx    = 8,  // I2S0_dev_rclk
  Clock_gate_hash       = 6,  // HASH
  Clock_gate_pwm        = 5,  // PWM
  Clock_gate_cdbus      = 2,  // CDBUS
  Clock_gate_can1       = 1,  // CAN1
  Clock_gate_can0       = 0,  // CAN0

  // Special value

  Clock_gate_undefined  = 32,
};

enum Clock_change_enable_bits : unsigned
{
  Clock_change_enable_cpu       = 22,
  Clock_change_enable_ahb0      = 21,
  Clock_change_enable_ahb2      = 20,
  Clock_change_enable_ddr       = 29,
  Clock_change_enable_mac       = 29,
  Clock_change_enable_i2s       = 29,
  Clock_change_enable_lcd       = 29,
  Clock_change_enable_msc0      = 29,
  Clock_change_enable_msc1      = 29,
  Clock_change_enable_sfc       = 29,
  Clock_change_enable_ssi       = 29,
  Clock_change_enable_cim       = 29,
  Clock_change_enable_pwm       = 29,
  Clock_change_enable_can0      = 29,
  Clock_change_enable_can1      = 29,
  Clock_change_enable_cdbus     = 29,

  // Special value

  Clock_change_enable_undefined = 32,
};

enum Clock_busy_bits : unsigned
{
  Clock_busy_cpu       = 0,
  Clock_busy_ddr       = 28,
  Clock_busy_mac       = 28,
  Clock_busy_lcd       = 28,
  Clock_busy_msc0      = 28,
  Clock_busy_msc1      = 28,
  Clock_busy_sfc       = 28,
  Clock_busy_ssi       = 28,
  Clock_busy_cim       = 28,
  Clock_busy_pwm       = 28,
  Clock_busy_can0      = 28,
  Clock_busy_can1      = 28,
  Clock_busy_cdbus     = 28,

  // Special value

  Clock_busy_undefined = 32,
};

enum Clock_divider_bits : unsigned
{
  Clock_divider_can0      = 0,  // CAN0CDR
  Clock_divider_can1      = 0,  // CAN1CDR
  Clock_divider_cdbus     = 0,  // CDBUSCDR
  Clock_divider_cim       = 0,  // CIMCDR
  Clock_divider_cpu       = 0,  // CDIV
  Clock_divider_ddr       = 0,  // DDRCDR
  Clock_divider_hclock0   = 8,  // H0DIV (fast AHB peripherals)
  Clock_divider_hclock2   = 12, // H2DIV (fast AHB peripherals)
  Clock_divider_l2cache   = 4,  // L2CDIV
  Clock_divider_lcd       = 0,  // LPCDR
  Clock_divider_mac       = 0,  // MACCDR
  Clock_divider_msc0      = 0,  // MSC0CDR
  Clock_divider_msc1      = 0,  // MSC1CDR
  Clock_divider_pclock    = 16, // PDIV (slow APB peripherals)
  Clock_divider_pwm       = 0,  // PWMCDR
  Clock_divider_sfc       = 0,  // SFCCDR
  Clock_divider_ssi       = 0,  // SSICDR

  // Special value

  Clock_divider_undefined = 32,
};

enum Pll_bits : unsigned
{
  // Pll_control_A, Pll_control_M, Pll_control_E

  Pll_multiplier        = 20, // xPLLM
  Pll_input_division    = 14, // xPLLN
  Pll_output_division1  = 11, // xPLLOD1
  Pll_output_division0  = 8,  // xPLLOD0
  Pll_stable            = 3,  // xPLL_ON
  Pll_enabled           = 0,  // xPLLEN
};

enum Pll_bypass_bits : unsigned
{
  Pll_bypass_A          = 30, // APLL_BP
  Pll_bypass_M          = 28, // MPLL_BP
  Pll_bypass_E          = 26, // EPLL_BP
};



// Clock input descriptions.

struct Clock_input_desc
{
  uint32_t source_reg;
  enum Clock_source_bits source_bit;
  int num_inputs;
  enum Clock_input_identifiers inputs[3];
};

struct Clock_input_desc clock_input_desc[Clock_input_identifier_count] = {

  /* Clock_input_ahb2_apb */  {Clock_control, Clock_source_hclock2,
                               3, {Clock_input_none, Clock_input_main, Clock_input_pll_M}},

  /* Clock_input_external */  {Reg_undefined, Clock_source_undefined,
                               0, {}},

  /* Clock_input_main */      {Clock_control, Clock_source_main,
                               3, {Clock_input_none, Clock_input_external, Clock_input_pll_A}},

  /* Clock_input_none */      {Reg_undefined, Clock_source_undefined,
                               0, {}},

  /* Clock_input_pll_A */     {Reg_undefined, Clock_source_undefined,
                               1, {Clock_input_external}},

  /* Clock_input_pll_E */     {Reg_undefined, Clock_source_undefined,
                               1, {Clock_input_external}},

  /* Clock_input_pll_M */     {Reg_undefined, Clock_source_undefined,
                               1, {Clock_input_external}},
};



// Clock descriptions.

struct Clock_desc
{
  uint32_t source_reg;
  enum Clock_source_bits source_bit;
  uint32_t gate_reg;
  enum Clock_gate_bits gate_bit;
  uint32_t change_enable_reg;
  enum Clock_change_enable_bits change_enable_bit;
  uint32_t busy_reg;
  enum Clock_busy_bits busy_bit;
  uint32_t divider_reg;
  enum Clock_divider_bits divider_bit;
  uint32_t divider_mask;
  int num_inputs;
  enum Clock_input_identifiers inputs[4];
  enum Clock_identifiers clock_input;
};

#define Clock_desc_undefined {Reg_undefined, Clock_source_undefined, \
                              Reg_undefined, Clock_gate_undefined, \
                              Reg_undefined, Clock_change_enable_undefined, \
                              Reg_undefined, Clock_busy_undefined, \
                              Reg_undefined, Clock_divider_undefined, 0, \
                              0, {}, \
                              Clock_undefined}

static struct Clock_desc clock_desc[Clock_identifier_count] = {

  /* Clock_aic_bitclk */  Clock_desc_undefined,

  /* Clock_aic_pclk */    Clock_desc_undefined,

  /* Clock_can0 */        {Can_divider0, Clock_source_can0,
                           Clock_gate1, Clock_gate_can0,
                           Can_divider0, Clock_change_enable_can0,
                           Can_divider0, Clock_busy_can0,
                           Can_divider0, Clock_divider_can0, 0xff,
                           4, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E, Clock_input_external},
                           Clock_undefined},

  /* Clock_can1 */        {Can_divider1, Clock_source_can1,
                           Clock_gate1, Clock_gate_can1,
                           Can_divider1, Clock_change_enable_can1,
                           Can_divider1, Clock_busy_can1,
                           Can_divider1, Clock_divider_can1, 0xff,
                           4, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E, Clock_input_external},
                           Clock_undefined},

  /* Clock_cdbus */       {Cdbus_divider, Clock_source_cdbus,
                           Clock_gate1, Clock_gate_cdbus,
                           Cdbus_divider, Clock_change_enable_cdbus,
                           Cdbus_divider, Clock_busy_cdbus,
                           Cdbus_divider, Clock_divider_cdbus, 0xff,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_cim */         {Cim_divider, Clock_source_cim,
                           Clock_gate0, Clock_gate_cim,
                           Cim_divider, Clock_change_enable_cim,
                           Cim_divider, Clock_busy_cim,
                           Cim_divider, Clock_divider_cim, 0xff,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_cpu */         {Clock_control, Clock_source_cpu,
                           Reg_undefined, Clock_gate_undefined,
                           Clock_control, Clock_change_enable_cpu,
                           Clock_status, Clock_busy_cpu,
                           Clock_control, Clock_divider_cpu, 0x0f,
                           3, {Clock_input_none, Clock_input_main, Clock_input_pll_M},
                           Clock_undefined},

  /* Clock_ddr */         {Ddr_divider, Clock_source_ddr,
                           Clock_gate0, Clock_gate_ddr,
                           Ddr_divider, Clock_change_enable_ddr,
                           Ddr_divider, Clock_busy_ddr,
                           Ddr_divider, Clock_divider_ddr, 0x0f,
                           3, {Clock_input_none, Clock_input_main, Clock_input_pll_M},
                           Clock_undefined},

  /* Clock_dma */         {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_dma,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           0, {},
                           Clock_pclock},

  /* Clock_emac */        Clock_desc_undefined,

  /* Clock_hclock0 */     {Clock_control, Clock_source_hclock0,
                           Clock_gate0, Clock_gate_ahb0,
                           Clock_control, Clock_change_enable_ahb0,
                           Reg_undefined, Clock_busy_undefined,
                           Clock_control, Clock_divider_hclock0, 0x0f,
                           3, {Clock_input_none, Clock_input_main, Clock_input_pll_M},
                           Clock_undefined},

  /* Clock_hclock2 */     {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_apb0,
                           Clock_control, Clock_change_enable_ahb2,
                           Reg_undefined, Clock_busy_undefined,
                           Clock_control, Clock_divider_hclock2, 0x0f,
                           1, {Clock_input_ahb2_apb},
                           Clock_undefined},

  /* Clock_hdmi */        Clock_desc_undefined,

  /* Clock_i2c */         {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_i2c0,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           0, {},
                           Clock_pclock},

  /* Clock_i2c0 */        {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_i2c0,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           0, {},
                           Clock_pclock},

  /* Clock_i2c1 */        {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_i2c1,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           0, {},
                           Clock_pclock},

  /* Clock_i2s */         Clock_desc_undefined,

  /* Clock_i2s0_rx */     {I2s_divider0, Clock_source_i2s,
                           Clock_gate1, Clock_gate_i2s0_rx,
                           I2s_divider0, Clock_change_enable_i2s,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0, // NOTE: To define.
                           2, {Clock_input_main, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_i2s0_tx */     {I2s_divider0, Clock_source_i2s,
                           Clock_gate1, Clock_gate_i2s0_tx,
                           I2s_divider0, Clock_change_enable_i2s,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0, // NOTE: To define.
                           2, {Clock_input_main, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_kbc */         Clock_desc_undefined,

  /* Clock_lcd */         Clock_desc_undefined,

  /* Clock_lcd_pixel */   {Lcd_divider, Clock_source_lcd,
                           Clock_gate0, Clock_gate_lcd_pixel,
                           Lcd_divider, Clock_change_enable_lcd,
                           Lcd_divider, Clock_busy_lcd,
                           Lcd_divider, Clock_divider_lcd, 0xff,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_mac */         {Mac_divider, Clock_source_mac,
                           Clock_gate1, Clock_gate_gmac0,
                           Mac_divider, Clock_change_enable_mac,
                           Mac_divider, Clock_busy_mac,
                           Mac_divider, Clock_divider_mac, 0xff,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_main */        {Reg_undefined, Clock_source_undefined,
                           Clock_control, Clock_gate_main,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           1, {Clock_input_main},
                           Clock_undefined},

  /* Clock_msc */         {Msc_divider0, Clock_source_msc0,
                           Clock_gate0, Clock_gate_msc0,
                           Msc_divider0, Clock_change_enable_msc0,
                           Msc_divider0, Clock_busy_msc0,
                           Msc_divider0, Clock_divider_msc0, 0xff,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_msc0 */        {Msc_divider0, Clock_source_msc0,
                           Clock_gate0, Clock_gate_msc0,
                           Msc_divider0, Clock_change_enable_msc0,
                           Msc_divider0, Clock_busy_msc0,
                           Msc_divider0, Clock_divider_msc0, 0xff,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_msc1 */        {Msc_divider1, Clock_source_msc1,
                           Clock_gate0, Clock_gate_msc1,
                           Msc_divider1, Clock_change_enable_msc1,
                           Msc_divider1, Clock_busy_msc1,
                           Msc_divider1, Clock_divider_msc1, 0xff,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_pclock */      {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_apb0,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Clock_control, Clock_divider_pclock, 0x0f,
                           1, {Clock_input_ahb2_apb},
                           Clock_undefined},

  /* Clock_pwm */         {Pwm_divider, Clock_source_pwm,
                           Clock_gate1, Clock_gate_pwm,
                           Pwm_divider, Clock_change_enable_pwm,
                           Pwm_divider, Clock_busy_pwm,
                           Pwm_divider, Clock_divider_pwm, 0x0f,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_pwm0 */        {Pwm_divider, Clock_source_pwm,
                           Clock_gate1, Clock_gate_pwm,
                           Pwm_divider, Clock_change_enable_pwm,
                           Pwm_divider, Clock_busy_pwm,
                           Pwm_divider, Clock_divider_pwm, 0x0f,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_pwm1 */        Clock_desc_undefined,

  /* Clock_scc */         Clock_desc_undefined,

  /* Clock_sfc */         {Sfc_divider, Clock_source_sfc,
                           Clock_gate0, Clock_gate_sfc,
                           Sfc_divider, Clock_change_enable_sfc,
                           Sfc_divider, Clock_busy_sfc,
                           Sfc_divider, Clock_divider_sfc, 0xff,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_smb0 */        Clock_desc_undefined,

  /* Clock_smb1 */        Clock_desc_undefined,

  /* Clock_smb2 */        Clock_desc_undefined,

  /* Clock_smb3 */        Clock_desc_undefined,

  /* Clock_smb4 */        Clock_desc_undefined,

  /* Clock_ssi */         {Ssi_divider, Clock_source_ssi,
                           Clock_gate0, Clock_gate_ssi0,
                           Ssi_divider, Clock_change_enable_ssi,
                           Ssi_divider, Clock_busy_ssi,
                           Ssi_divider, Clock_divider_ssi, 0xff,
                           3, {Clock_input_main, Clock_input_pll_M, Clock_input_pll_E},
                           Clock_undefined},

  /* Clock_timer */       {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_timer,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           0, {},
                           Clock_pclock},

  /* Clock_uart0 */       {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_uart0,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           1, {Clock_input_external},
                           Clock_undefined},

  /* Clock_uart1 */       {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_uart1,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           1, {Clock_input_external},
                           Clock_undefined},

  /* Clock_uart2 */       {Reg_undefined, Clock_source_undefined,
                           Clock_gate0, Clock_gate_uart2,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           1, {Clock_input_external},
                           Clock_undefined},

  /* Clock_uart3 */       {Reg_undefined, Clock_source_undefined,
                           Clock_gate1, Clock_gate_uart3,
                           Reg_undefined, Clock_change_enable_undefined,
                           Reg_undefined, Clock_busy_undefined,
                           Reg_undefined, Clock_divider_undefined, 0,
                           1, {Clock_input_external},
                           Clock_undefined},

  /* Clock_udc */         Clock_desc_undefined,

  /* Clock_uhc */         Clock_desc_undefined,

  /* Clock_uprt */        Clock_desc_undefined,
};



// Convenience functions.

static uint8_t get_clock_gate_bit(enum Clock_identifiers clock)
{
  enum Clock_gate_bits bit = clock_desc[clock].gate_bit;

  return bit != Clock_gate_undefined ? (uint8_t) bit : 0;
}

static uint32_t get_clock_gate_mask(enum Clock_identifiers clock)
{
  enum Clock_gate_bits bit = clock_desc[clock].gate_bit;

  return bit != Clock_gate_undefined ? 1 : 0;
}



// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Cpm_x1600_chip::Cpm_x1600_chip(l4_addr_t addr, uint32_t exclk_freq)
: _exclk_freq(exclk_freq)
{
  _regs = new Hw::Mmio_register_block<32>(addr);

  // add_cid("cpm");
  // add_cid("cpm-x1600");
  // register_property("exclk_freq", &_exclk_freq);
}



// Utility methods.

uint32_t
Cpm_x1600_chip::get_field(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return (_regs[reg] & (mask << shift)) >> shift;
}

void
Cpm_x1600_chip::set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value)
{
  _regs[reg] = (_regs[reg] & (~(mask << shift))) | ((mask & value) << shift);
}



// Clock/timer control.

void
Cpm_x1600_chip::change_disable(enum Clock_identifiers clock)
{
  enum Clock_change_enable_bits bit = clock_desc[clock].change_enable_bit;

  if (bit != Clock_change_enable_undefined)
    set_field(clock_desc[clock].change_enable_reg, 1, bit, 0);
}

void
Cpm_x1600_chip::change_enable(enum Clock_identifiers clock)
{
  enum Clock_change_enable_bits bit = clock_desc[clock].change_enable_bit;

  if (bit != Clock_change_enable_undefined)
    set_field(clock_desc[clock].change_enable_reg, 1, bit, 1);
}

int
Cpm_x1600_chip::have_clock(enum Clock_identifiers clock)
{
  if (clock_desc[clock].gate_bit != Clock_gate_undefined)
    return !get_field(clock_desc[clock].gate_reg, get_clock_gate_mask(clock),
                      get_clock_gate_bit(clock));
  else
    return true;
}

void
Cpm_x1600_chip::start_clock(enum Clock_identifiers clock)
{
  if (clock_desc[clock].gate_bit != Clock_gate_undefined)
    set_field(clock_desc[clock].gate_reg, get_clock_gate_mask(clock),
              get_clock_gate_bit(clock), 0);
}

void
Cpm_x1600_chip::stop_clock(enum Clock_identifiers clock)
{
  if (clock_desc[clock].gate_bit != Clock_gate_undefined)
    set_field(clock_desc[clock].gate_reg, get_clock_gate_mask(clock),
              get_clock_gate_bit(clock), 1);
}

void
Cpm_x1600_chip::wait_busy(enum Clock_identifiers clock)
{
  enum Clock_busy_bits bit = clock_desc[clock].busy_bit;

  if (bit != Clock_busy_undefined)
    while (get_field(clock_desc[clock].busy_reg, 1, bit));
}



// PLL control.

// Return whether the PLL is stable.

int
Cpm_x1600_chip::have_pll(uint32_t pll_reg)
{
  return get_field(pll_reg, 1, Pll_stable);
}

int
Cpm_x1600_chip::pll_enabled(uint32_t pll_reg)
{
  return get_field(pll_reg, 1, Pll_enabled);
}

int
Cpm_x1600_chip::pll_bypassed(uint32_t pll_reg)
{
  uint8_t bit;
  unsigned mask = 1;

  switch (pll_reg)
  {
    case Pll_control_A: bit = Pll_bypass_A; break;
    case Pll_control_M: bit = Pll_bypass_M; break;
    case Pll_control_E: bit = Pll_bypass_E; break;
    default: bit = 0; mask = 0; break;
  }

  return get_field(Pll_control, mask, bit);
}

void
Cpm_x1600_chip::pll_enable(uint32_t pll_reg)
{
  set_field(pll_reg, 1, Pll_enabled, 1);
  while (!have_pll(pll_reg));
}

void
Cpm_x1600_chip::pll_disable(uint32_t pll_reg)
{
  set_field(pll_reg, 1, Pll_enabled, 0);
  while (have_pll(pll_reg));
}

// Feedback (13-bit) multiplier.

uint16_t
Cpm_x1600_chip::get_multiplier(uint32_t pll_reg)
{
  return get_field(pll_reg, 0x1fff, Pll_multiplier) + 1;
}

void
Cpm_x1600_chip::set_multiplier(uint32_t pll_reg, uint16_t multiplier)
{
  set_field(pll_reg, 0x1fff, Pll_multiplier, multiplier - 1);
}

// Input (6-bit) divider.

uint8_t
Cpm_x1600_chip::get_input_division(uint32_t pll_reg)
{
  return get_field(pll_reg, 0x3f, Pll_input_division) + 1;
}

void
Cpm_x1600_chip::set_input_division(uint32_t pll_reg, uint8_t divider)
{
  set_field(pll_reg, 0x3f, Pll_input_division, divider - 1);
}

// Output (dual 3-bit) dividers.

uint8_t
Cpm_x1600_chip::get_output_division(uint32_t pll_reg)
{
  uint8_t d0 = get_field(pll_reg, 0x07, Pll_output_division0);
  uint8_t d1 = get_field(pll_reg, 0x07, Pll_output_division1);

  return d0 * d1;
}

void
Cpm_x1600_chip::set_output_division(uint32_t pll_reg, uint8_t divider)
{
  // Assert 1 as a minimum.
  // Divider 0 must be less than or equal to divider 1.

  uint8_t d0 = (uint8_t) floor(sqrt(divider ? divider : 1));
  uint8_t d1 = divider / d0;

  set_field(pll_reg, 0x07, Pll_output_division0, d0);
  set_field(pll_reg, 0x07, Pll_output_division1, d1);
}

uint32_t
Cpm_x1600_chip::get_pll_frequency(uint32_t pll_reg)
{
  // Test for PLL enable and not PLL bypass.

  if (pll_enabled(pll_reg) && !pll_bypassed(pll_reg))
    return (_exclk_freq * get_multiplier(pll_reg)) /
           (get_input_division(pll_reg) * get_output_division(pll_reg));
  else
    return _exclk_freq;
}

void
Cpm_x1600_chip::set_pll_parameters(uint32_t pll_reg, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  set_multiplier(pll_reg, multiplier);
  set_input_division(pll_reg, in_divider);
  set_output_division(pll_reg, out_divider);

  if (pll_enabled(pll_reg) && !pll_bypassed(pll_reg))
    while (!have_pll(pll_reg));
}



// Clock dividers.

uint32_t
Cpm_x1600_chip::get_divider(enum Clock_identifiers clock)
{
  if (clock_desc[clock].divider_bit != Clock_divider_undefined)
    return get_field(clock_desc[clock].divider_reg, clock_desc[clock].divider_mask,
                     clock_desc[clock].divider_bit) + 1;
  else
    return 1;
}

void
Cpm_x1600_chip::set_divider(enum Clock_identifiers clock, uint32_t division)
{
  if (clock_desc[clock].divider_bit == Clock_divider_undefined)
    return;

  change_enable(clock);
  set_field(clock_desc[clock].divider_reg, clock_desc[clock].divider_mask,
            clock_desc[clock].divider_bit, division - 1);
  wait_busy(clock);
  change_disable(clock);
}



// Clock sources.

uint8_t
Cpm_x1600_chip::get_source(enum Clock_identifiers clock)
{
  if (clock_desc[clock].source_bit != Clock_source_undefined)
    return get_field(clock_desc[clock].source_reg, Source_mask, clock_desc[clock].source_bit);
  else
    return 0;
}

void
Cpm_x1600_chip::set_source(enum Clock_identifiers clock, uint8_t source)
{
  if (clock_desc[clock].source_bit == Clock_source_undefined)
    return;

  change_enable(clock);
  set_field(clock_desc[clock].source_reg, Source_mask, clock_desc[clock].source_bit, source);
  wait_busy(clock);
  change_disable(clock);
}



// Clock source frequencies.

uint32_t
Cpm_x1600_chip::get_input_frequency(enum Clock_input_identifiers clock)
{
  struct Clock_input_desc desc = clock_input_desc[clock];

  // Clocks with no inputs provide a frequency.

  if (desc.num_inputs == 0)
  {
    switch (clock)
    {
      case Clock_input_external: return _exclk_freq;
      default: return 0;
    }
  }

  // Of the input clocks, only PLLs have a single input.

  else if (desc.num_inputs == 1)
  {
    switch (clock)
    {
      case Clock_input_pll_A: return get_pll_frequency(Pll_control_A);
      case Clock_input_pll_E: return get_pll_frequency(Pll_control_E);
      case Clock_input_pll_M: return get_pll_frequency(Pll_control_M);
      default: return 0;
    }
  }

  // With multiple sources, obtain the selected source for the clock.

  uint8_t source = get_field(desc.source_reg, Source_mask, desc.source_bit);

  // Return the frequency of the source.

  if (source < desc.num_inputs)
    return get_input_frequency(desc.inputs[source]);
  else
    return 0;
}

uint32_t
Cpm_x1600_chip::get_source_frequency(enum Clock_identifiers clock)
{
  struct Clock_desc desc = clock_desc[clock];

  if (desc.num_inputs == 0)
  {
    // Clocks may reference other clocks.

    if (desc.clock_input != Clock_undefined)
      return get_frequency(desc.clock_input);

    // Undefined clocks return zero.

    else
      return 0;
  }

  // Clocks with one source yield that input frequency.

  else if (desc.num_inputs == 1)
    return get_input_frequency(desc.inputs[0]);

  // With multiple sources, obtain the selected source for the clock.

  uint8_t source = get_source(clock);

  // Return the frequency of the source.

  if (source < desc.num_inputs)
    return get_input_frequency(desc.inputs[source]);
  else
    return 0;
}



// Output clock frequencies.

uint32_t
Cpm_x1600_chip::get_frequency(enum Clock_identifiers clock)
{
  return get_source_frequency(clock) / get_divider(clock);
}

void
Cpm_x1600_chip::set_frequency(enum Clock_identifiers clock, uint32_t frequency)
{
  switch (clock)
  {
    // The pixel frequency is based on the selected clock source (SCLK_A, MPLL or
    // EPLL).

    case Clock_lcd_pixel:

      // Switch to the MPLL and attempt to set the divider.

      set_source(Clock_lcd_pixel, Source_mME_pll_M);
      pll_enable(Pll_control_M);
      set_divider(Clock_lcd_pixel, get_source_frequency(clock) / frequency);
      break;

    default:
      break;
  }
}



// C language interface functions.

void
*x1600_cpm_init(l4_addr_t cpm_base)
{
  /* Initialise the clock and power management peripheral with the
     register memory region and a 24MHz EXCLK frequency. */

  return (void *) new Cpm_x1600_chip(cpm_base, 24000000);
}

int
x1600_cpm_have_clock(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->have_clock(clock);
}

void
x1600_cpm_start_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->start_clock(clock);
}

void
x1600_cpm_stop_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->stop_clock(clock);
}



uint32_t
x1600_cpm_get_divider(void *cpm, enum Clock_identifiers clock)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_divider(clock);
}

void
x1600_cpm_set_divider(void *cpm, enum Clock_identifiers clock, uint32_t divider)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->set_divider(clock, divider);
}



uint8_t
x1600_cpm_get_source(void *cpm, enum Clock_identifiers clock)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_source(clock);
}

void
x1600_cpm_set_source(void *cpm, enum Clock_identifiers clock, uint8_t source)
{       
  static_cast<Cpm_x1600_chip *>(cpm)->set_source(clock, source);
}



uint32_t
x1600_cpm_get_source_frequency(void *cpm, enum Clock_identifiers clock)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_source_frequency(clock);
}



uint32_t
x1600_cpm_get_frequency(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_frequency(clock);
}

void
x1600_cpm_set_frequency(void *cpm, enum Clock_identifiers clock, uint32_t frequency)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_frequency(clock, frequency);
}



void
x1600_cpm_set_mpll_parameters(void *cpm, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_pll_parameters(Pll_control_M, multiplier, in_divider, out_divider);
}
